@@ -25,27 +25,31 @@
  * Corel Corporation or Corel Corporation Limited."
  */
 
-#define DUMP_BINARY_DATA 0
-
-#include <librevenge/librevenge.h>
-#include <libwpd/libwpd.h>
 #include "WPG2Parser.h"
-#include "libwpg_utils.h"
 
 #include <cstring>
 #include <limits>
 #include <math.h>
 #include <vector>
 
-// MSVC++ 6.0 does not have the macro defined, so we define it
-#ifndef M_PI
-#define M_PI 3.14159265358979323846
-#endif
+#include <librevenge/librevenge.h>
+
+#include <libwpd/libwpd.h>
+
+#include "WPGTextDataHandler.h"
+#include "libwpg_utils.h"
+
+#define DUMP_BINARY_DATA 0
 
 #if DUMP_BINARY_DATA
 #include <sstream>
 #endif
 
+// MSVC++ 6.0 does not have the macro defined, so we define it
+#ifndef M_PI
+#define M_PI 3.14159265358979323846
+#endif
+
 using namespace libwpd;
 
 using libwpg::getRemainingLength;
@@ -55,12 +59,32 @@ namespace
 
 static double fixedPointToDouble(const unsigned int fixedPointNumber)
 {
-	short fixedPointNumberIntegerPart = (short)((fixedPointNumber & 0xFFFF0000) >> 16);
-	double fixedPointNumberFractionalPart = (double)((double)(fixedPointNumber & 0x0000FFFF)/(double)0xFFFF);
+	auto fixedPointNumberIntegerPart = (short)((fixedPointNumber & 0xFFFF0000) >> 16);
+	auto fixedPointNumberFractionalPart = (double)((double)(fixedPointNumber & 0x0000FFFF)/(double)0xFFFF);
 	return ((fixedPointNumberIntegerPart >= 0) ?
 	        ((double)fixedPointNumberIntegerPart + fixedPointNumberFractionalPart) :
 	        ((double)fixedPointNumberIntegerPart - fixedPointNumberFractionalPart));
 }
+
+long safeSubtract(long left, const long right)
+{
+	if (right > 0)
+	{
+		if (left > std::numeric_limits<long>::min() + right)
+			left -= right;
+		else
+			left = std::numeric_limits<long>::min();
+	}
+	else if (right < 0)
+	{
+		if (left < std::numeric_limits<long>::max() + right)
+			left -= right;
+		else
+			left = std::numeric_limits<long>::max();
+	}
+	return left;
+}
+
 }
 
 static const unsigned char defaultWPG2PaletteRed[] =
@@ -273,13 +297,13 @@ WPG2Parser::WPG2Parser(librevenge::RVNGInputStream *input, librevenge::RVNGDrawi
 {
 	m_style.insert("draw:fill", "solid");
 	// default style
-	m_penForeColor = ::libwpg::WPGColor(0,0,0);
-	m_penBackColor = ::libwpg::WPGColor(0xff,0xff,0xff);
+	m_penForeColor = libwpg::WPGColor(0,0,0);
+	m_penBackColor = libwpg::WPGColor(0xff,0xff,0xff);
 	m_style.insert("svg:stroke-width", 0.0);
 	m_style.insert("draw:stroke", "solid");
 	m_dashArray = libwpg::WPGDashArray();
-	m_brushForeColor = ::libwpg::WPGColor(0,0,0);
-	m_brushBackColor = ::libwpg::WPGColor(0xff,0xff,0xff);
+	m_brushForeColor = libwpg::WPGColor(0,0,0);
+	m_brushBackColor = libwpg::WPGColor(0xff,0xff,0xff);
 
 	m_style.insert("svg:stroke-color", m_penForeColor.getColorString());
 	m_style.insert("svg:stroke-opacity", m_penForeColor.getOpacity(), librevenge::RVNG_PERCENT);
@@ -306,19 +330,19 @@ bool WPG2Parser::parse()
 		{ 0x01, "Start WPG",            &WPG2Parser::handleStartWPG },
 		{ 0x02, "End WPG",              &WPG2Parser::handleEndWPG },
 		{ 0x03, "Form Settings",        &WPG2Parser::handleFormSettings },
-		{ 0x04, "Ruler Settings",       0 },     // ignored
-		{ 0x05, "Grid Settings",        0 },     // ignored
+		{ 0x04, "Ruler Settings",       nullptr },     // ignored
+		{ 0x05, "Grid Settings",        nullptr },     // ignored
 		{ 0x06, "Layer",                &WPG2Parser::handleLayer },
 		{ 0x08, "Pen Style Definition", &WPG2Parser::handlePenStyleDefinition },
-		{ 0x09, "Pattern Definition",   0 },
-		{ 0x0a, "Comment",              0 },     // ignored
-		{ 0x0b, "Color Transfer",       0 },
+		{ 0x09, "Pattern Definition",   &WPG2Parser::handlePatternDefinition },
+		{ 0x0a, "Comment",              nullptr },     // ignored
+		{ 0x0b, "Color Transfer",       nullptr },
 		{ 0x0c, "Color Palette",        &WPG2Parser::handleColorPalette },
 		{ 0x0d, "DP Color Palette",     &WPG2Parser::handleDPColorPalette },
 		{ 0x0e, "Bitmap Data",          &WPG2Parser::handleBitmapData },
 		{ 0x0f, "Text Data",            &WPG2Parser::handleTextData },
-		{ 0x10, "Chart Style",          0 },     // ignored
-		{ 0x11, "Chart Data",           0 },     // ignored
+		{ 0x10, "Chart Style",          nullptr },     // ignored
+		{ 0x11, "Chart Data",           nullptr },     // ignored
 		{ 0x12, "Object Image",         &WPG2Parser::handleObjectImage },
 		{ 0x15, "Polyline",             &WPG2Parser::handlePolyline },
 		{ 0x16, "Polyspline",           &WPG2Parser::handlePolyspline },
@@ -330,16 +354,16 @@ bool WPG2Parser::parse()
 		{ 0x1c, "Text Line",            &WPG2Parser::handleTextLine },
 		{ 0x1d, "Text Block",           &WPG2Parser::handleTextBlock },
 		{ 0x1e, "Text Path",            &WPG2Parser::handleTextPath },
-		{ 0x1f, "Chart",                0 },
-		{ 0x20, "Group",                0 },
+		{ 0x1f, "Chart",                nullptr },
+		{ 0x20, "Group",                nullptr },
 		{ 0x21, "Object Capsule",       &WPG2Parser::handleObjectCapsule },
-		{ 0x22, "Font Settings",        0 },
+		{ 0x22, "Font Settings",        nullptr },
 		{ 0x25, "Pen Fore Color",       &WPG2Parser::handlePenForeColor },
 		{ 0x26, "DP Pen Fore Color",    &WPG2Parser::handleDPPenForeColor },
 		{ 0x27, "Pen Back Color",       &WPG2Parser::handlePenBackColor },
 		{ 0x28, "DP Pen Back Color",    &WPG2Parser::handleDPPenBackColor },
 		{ 0x29, "Pen Style",            &WPG2Parser::handlePenStyle },
-		{ 0x2a, "Pen Pattern",          0 },
+		{ 0x2a, "Pen Pattern",          nullptr },
 		{ 0x2b, "Pen Size",             &WPG2Parser::handlePenSize },
 		{ 0x2c, "DP Pen Size",          &WPG2Parser::handleDPPenSize  },
 		{ 0x2d, "Line Cap",             &WPG2Parser::handleLineCap },
@@ -351,17 +375,17 @@ bool WPG2Parser::parse()
 		{ 0x33, "Brush Back Color",     &WPG2Parser::handleBrushBackColor },
 		{ 0x34, "DP Brush Back Color",  &WPG2Parser::handleDPBrushBackColor },
 		{ 0x35, "Brush Pattern",        &WPG2Parser::handleBrushPattern },
-		{ 0x36, "Horizontal Line",      0 },
-		{ 0x37, "Vertical Line",        0 },
-		{ 0x38, "Poster Settings",      0 },
-		{ 0x39, "Image State",          0 },
-		{ 0x3a, "Envelope Definition",  0 },
-		{ 0x3b, "Envelope",             0 },
-		{ 0x3c, "Texture Definition",   0 },
-		{ 0x3d, "Brush Texture",        0 },
-		{ 0x3e, "Texture Alignment",    0 },
-		{ 0x3f, "Pen Texture ",         0 },
-		{ 0x00, 0, 0 } // end marker
+		{ 0x36, "Horizontal Line",      nullptr },
+		{ 0x37, "Vertical Line",        nullptr },
+		{ 0x38, "Poster Settings",      nullptr },
+		{ 0x39, "Image State",          nullptr },
+		{ 0x3a, "Envelope Definition",  nullptr },
+		{ 0x3b, "Envelope",             nullptr },
+		{ 0x3c, "Texture Definition",   nullptr },
+		{ 0x3d, "Brush Texture",        nullptr },
+		{ 0x3e, "Texture Alignment",    nullptr },
+		{ 0x3f, "Pen Texture ",         nullptr },
+		{ 0x00, nullptr, nullptr } // end marker
 	};
 
 	while (!m_input->isEnd())
@@ -373,7 +397,7 @@ bool WPG2Parser::parse()
 		int recordType = readU8();
 		if (recordType == 0 || recordType > (int)0x3f)
 			break;
-		int extension = (int) readVariableLengthInteger();
+		auto extension = (int) readVariableLengthInteger();
 		unsigned int recordLength = readVariableLengthInteger();
 		const unsigned long maxLength = getRemainingLength(m_input);
 		if (recordLength > maxLength)
@@ -527,9 +551,6 @@ static const char *describeGradient(unsigned char gradientType)
 }
 #endif
 
-#define TO_DOUBLE(x) ( (m_doublePrecision) ? ((double)(x)/65536.0) : (double)(x) )
-#define TRANSFORM_XY(x,y) { m_matrix.transform((x),(y)); (x)-= m_xofs; (y)-= m_yofs; (y)=m_height-(y); }
-
 void WPG2Parser::handleStartWPG()
 {
 	if (m_graphicsStarted) // second start wpg in the document == possible corruption, bail out
@@ -596,11 +617,11 @@ void WPG2Parser::handleStartWPG()
 	WPG_DEBUG_MSG(("                       width : %li\n", m_width));
 	WPG_DEBUG_MSG(("                      height : %li\n", m_height));
 
-	::librevenge::RVNGPropertyList propList;
-	propList.insert("svg:width", ((TO_DOUBLE(m_width)) / m_xres));
-	propList.insert("svg:height", ((TO_DOUBLE(m_height)) / m_yres));
+	librevenge::RVNGPropertyList propList;
+	propList.insert("svg:width", ((toDouble(m_width)) / m_xres));
+	propList.insert("svg:height", ((toDouble(m_height)) / m_yres));
 
-	m_painter->startDocument(::librevenge::RVNGPropertyList());
+	m_painter->startDocument(librevenge::RVNGPropertyList());
 	m_painter->startPage(propList);
 
 	static const int WPG2_defaultPenDashes[] =
@@ -655,17 +676,17 @@ void WPG2Parser::handleFormSettings()
 #ifdef DEBUG
 	unsigned int w = (m_doublePrecision) ? readU32() : readU16();
 	unsigned int h = (m_doublePrecision) ? readU32() : readU16();
-	double width = (TO_DOUBLE(w)) / m_xres;
-	double height = (TO_DOUBLE(h)) / m_yres;
+	double width = (toDouble(w)) / m_xres;
+	double height = (toDouble(h)) / m_yres;
 	m_input->seek(((m_doublePrecision) ? 4 : 2), librevenge::RVNG_SEEK_CUR);
 	unsigned int ml = (m_doublePrecision) ? readU32() : readU16();
 	unsigned int mr = (m_doublePrecision) ? readU32() : readU16();
 	unsigned int mt = (m_doublePrecision) ? readU32() : readU16();
 	unsigned int mb = (m_doublePrecision) ? readU32() : readU16();
-	double margL = (TO_DOUBLE(ml)) / m_xres;
-	double margR = (TO_DOUBLE(mr)) / m_xres;
-	double margT = (TO_DOUBLE(mt)) / m_xres;
-	double margB = (TO_DOUBLE(mb)) / m_xres;
+	double margL = (toDouble(ml)) / m_xres;
+	double margR = (toDouble(mr)) / m_xres;
+	double margT = (toDouble(mt)) / m_xres;
+	double margB = (toDouble(mb)) / m_xres;
 
 	WPG_DEBUG_MSG(("Form Settings: width: %f height : %f\n", width, height));
 	WPG_DEBUG_MSG(("Form Margins:  left: %f right : %f top: %f bottom: %f\n", margL, margR, margT, margB));
@@ -752,8 +773,8 @@ void WPG2Parser::handlePenStyleDefinition()
 	{
 		unsigned int p = (m_doublePrecision) ? readU32() : readU16();
 		unsigned int q = (m_doublePrecision) ? readU32() : readU16();
-		dashArray.add(TO_DOUBLE(p)*3.6/218.0);
-		dashArray.add(TO_DOUBLE(q)*3.6/218.0);
+		dashArray.add(toDouble(p)*3.6/218.0);
+		dashArray.add(toDouble(q)*3.6/218.0);
 	}
 	m_dashArrayStyles[style] = dashArray;
 
@@ -761,7 +782,6 @@ void WPG2Parser::handlePenStyleDefinition()
 	WPG_DEBUG_MSG(("  Segment pairs : %d\n", (int) segments));
 }
 
-#if 0
 // TODO
 void WPG2Parser::handlePatternDefinition()
 {
@@ -769,7 +789,6 @@ void WPG2Parser::handlePatternDefinition()
 		return;
 	WPG_DEBUG_MSG(("PatternDefinition\n"));
 }
-#endif
 
 void WPG2Parser::handleColorPalette()
 {
@@ -827,6 +846,9 @@ void WPG2Parser::handlePenForeColor()
 
 	m_penForeColor = libwpg::WPGColor(red, green, blue, alpha);
 
+	m_style.insert("svg:stroke-color", m_penForeColor.getColorString());
+	m_style.insert("svg:stroke-opacity", m_penForeColor.getOpacity(), librevenge::RVNG_PERCENT);
+
 	WPG_DEBUG_MSG(("   Foreground color (RGBA): %d %d %d %d\n", red, green, blue, alpha));
 }
 
@@ -847,10 +869,11 @@ void WPG2Parser::handleDPPenForeColor()
 	unsigned char blue = (m_doublePrecision)  ? readU16()>>8 : readU8();
 	unsigned char alpha = 0xff - ((m_doublePrecision) ? readU16()>>8 : readU8());
 
-	m_style.insert("svg:stroke-color", libwpg::WPGColor(red, green, blue, alpha).getColorString());
-	m_style.insert("svg:stroke-opacity", libwpg::WPGColor(red, green, blue, alpha).getOpacity(), librevenge::RVNG_PERCENT);
 	m_penForeColor = libwpg::WPGColor(red, green, blue, alpha);
 
+	m_style.insert("svg:stroke-color", m_penForeColor.getColorString());
+	m_style.insert("svg:stroke-opacity", m_penForeColor.getOpacity(), librevenge::RVNG_PERCENT);
+
 	WPG_DEBUG_MSG(("   Foreground color (RGBA): %d %d %d %d\n", red, green, blue, alpha));
 }
 
@@ -947,7 +970,7 @@ void WPG2Parser::handlePenSize()
 	}
 	unsigned int width = readU16();
 
-	m_style.insert("svg:stroke-width", (TO_DOUBLE(width) / m_xres));
+	m_style.insert("svg:stroke-width", (toDouble(width) / m_xres));
 
 	WPG_DEBUG_MSG(("   Width: %d\n", (int) width));
 }
@@ -965,8 +988,8 @@ void WPG2Parser::handleDPPenSize()
 	}
 	unsigned long width = readU32();
 
-	m_style.insert("svg:stroke-width", TO_DOUBLE(width) / m_xres / 256);
-//	m_pen.height = TO_DOUBLE(height) / m_yres / 256;
+	m_style.insert("svg:stroke-width", toDouble(width) / m_xres / 256);
+//	m_pen.height = toDouble(height) / m_yres / 256;
 
 	WPG_DEBUG_MSG(("   Width: %li\n", (long) width));
 }
@@ -1122,7 +1145,7 @@ void WPG2Parser::handleBrushForeColor()
 			for (unsigned j = 0; j < count-1; j++)
 			{
 				unsigned pos = readU16();
-				positions.push_back(TO_DOUBLE(pos));
+				positions.push_back(toDouble(pos));
 				WPG_DEBUG_MSG(("   Position #%d : %d\n", j+1, (int) pos));
 			}
 		}
@@ -1138,7 +1161,7 @@ void WPG2Parser::handleBrushForeColor()
 			double angle = m_gradientAngle*M_PI/180.0;
 			double tanangle = tan(angle);
 			double ref = (tanangle < 1e2 && tanangle > -1e2) ? (yref+xref*tanangle)/(1+tanangle) : xref;
-			::librevenge::RVNGPropertyListVector gradient;
+			librevenge::RVNGPropertyListVector gradient;
 			m_style.insert("draw:angle", (int)(-m_gradientAngle)); // upside down
 			librevenge::RVNGPropertyList propList;
 			propList.insert("svg:offset", 0.0, librevenge::RVNG_PERCENT);
@@ -1217,7 +1240,7 @@ void WPG2Parser::handleDPBrushForeColor()
 			for (unsigned j = 0; j < count-1; j++)
 			{
 				unsigned pos = readU16();
-				positions.push_back(TO_DOUBLE(pos));
+				positions.push_back(toDouble(pos));
 				WPG_DEBUG_MSG(("   Position #%d : %d\n", j+1, (int) pos));
 			}
 		}
@@ -1233,7 +1256,7 @@ void WPG2Parser::handleDPBrushForeColor()
 			double angle = m_gradientAngle*M_PI/180.0;
 			double tanangle = tan(angle);
 			double ref = (tanangle<1e2) ? (yref+xref*tanangle)/(1+tanangle) : xref;
-			::librevenge::RVNGPropertyListVector gradient;
+			librevenge::RVNGPropertyListVector gradient;
 			m_style.insert("draw:angle", (int)(-m_gradientAngle));
 
 			librevenge::RVNGPropertyList propList;
@@ -1266,7 +1289,6 @@ void WPG2Parser::handleBrushBackColor()
 		return;
 	if (!m_groupStack.empty() && m_groupStack.top().isCompoundPolygon())
 		return;
-#if 0
 	unsigned char red = readU8();
 	unsigned char green = readU8();
 	unsigned char blue = readU8();
@@ -1274,12 +1296,13 @@ void WPG2Parser::handleBrushBackColor()
 
 	m_brushBackColor = libwpg::WPGColor(red, green, blue, alpha);
 
-	m_gradient.backColor = libwpg::WPGColor(red, green, blue, alpha);
-	if (m_gradient.style == libwpg::WPGGradient::NoBrush)
-		m_gradient.style = libwpg::WPGGradient::Solid;
+	m_style.insert("draw:fill-color", m_brushForeColor.getColorString());
+	m_style.insert("draw:opacity", m_brushForeColor.getOpacity(), librevenge::RVNG_PERCENT);
+
+	if (m_style["draw:fill"] && m_style["draw:fill"]->getStr() == "none")
+		m_style.insert("draw:fill", "solid");
 
 	WPG_DEBUG_MSG(("   Backround color (RGBA): %d %d %d %d\n", red, green, blue, alpha));
-#endif
 }
 
 void WPG2Parser::handleDPBrushBackColor()
@@ -1288,7 +1311,6 @@ void WPG2Parser::handleDPBrushBackColor()
 		return;
 	if (!m_groupStack.empty() && m_groupStack.top().isCompoundPolygon())
 		return;
-#if 0
 	// we just ignore the least significant 8 bits
 	unsigned int red = (m_doublePrecision)   ? readU16()>>8 : readU8();
 	unsigned int green = (m_doublePrecision) ? readU16()>>8 : readU8();
@@ -1304,7 +1326,6 @@ void WPG2Parser::handleDPBrushBackColor()
 		m_style.insert("draw:fill", "solid");
 
 	WPG_DEBUG_MSG(("   Background color (RGBA): %d %d %d %d\n", red, green, blue, alpha));
-#endif
 }
 
 void WPG2Parser::handleBrushPattern()
@@ -1440,19 +1461,19 @@ void WPG2Parser::handlePolyline()
 	if (count > maxCount)
 		count = maxCount;
 
-	::librevenge::RVNGPropertyListVector points;
+	librevenge::RVNGPropertyListVector points;
 	for (unsigned long i = 0; i < count; i++)
 	{
-		::librevenge::RVNGPropertyList point;
+		librevenge::RVNGPropertyList point;
 		long x = (m_doublePrecision) ? readS32() : readS16();
 		long y = (m_doublePrecision) ? readS32() : readS16();
-		TRANSFORM_XY(x,y);
-		point.insert("svg:x", (TO_DOUBLE(x)/m_xres));
-		point.insert("svg:y", (TO_DOUBLE(y)/m_yres));
+		transformXY(x,y);
+		point.insert("svg:x", (toDouble(x)/m_xres));
+		point.insert("svg:y", (toDouble(y)/m_yres));
 		points.append(point);
 	}
 
-	::librevenge::RVNGPropertyList tmpPoints;
+	librevenge::RVNGPropertyList tmpPoints;
 	tmpPoints.insert("svg:points", points);
 
 	if (insideCompound)
@@ -1461,8 +1482,8 @@ void WPG2Parser::handlePolyline()
 		{
 			// inside a compound ? convert it into path because for compound
 			// we will only use paths
-			::librevenge::RVNGPropertyListVector &path = m_groupStack.top().compoundPath;
-			::librevenge::RVNGPropertyList element;
+			librevenge::RVNGPropertyListVector &path = m_groupStack.top().compoundPath;
+			librevenge::RVNGPropertyList element;
 			element = points[0];
 			element.insert("librevenge:path-action", "M");
 			path.append(element);
@@ -1503,8 +1524,8 @@ void WPG2Parser::handlePolyline()
 	}
 
 	WPG_DEBUG_MSG(("   Vertices count : %li\n", (long) count));
-	for (unsigned int j = 0; j < count; j++)
-		WPG_DEBUG_MSG(("        Point #%d : %g,%g\n", j+1, points[j]["svg:x"]->getDouble(), points[j]["svg:x"]->getDouble()));
+	for (unsigned long j = 0; j < count; j++)
+		WPG_DEBUG_MSG(("        Point #%lu : %g,%g\n", j+1, points[j]["svg:x"]->getDouble(), points[j]["svg:x"]->getDouble()));
 }
 
 void WPG2Parser::handlePolyspline()
@@ -1547,37 +1568,37 @@ void WPG2Parser::handlePolycurve()
 	if (count > maxCount)
 		count = maxCount;
 
-	::librevenge::RVNGPropertyListVector path;
-	::librevenge::RVNGPropertyList element;
-	::librevenge::RVNGPropertyListVector vertices;
-	::librevenge::RVNGPropertyListVector controlPoints;
+	librevenge::RVNGPropertyListVector path;
+	librevenge::RVNGPropertyList element;
+	librevenge::RVNGPropertyListVector vertices;
+	librevenge::RVNGPropertyListVector controlPoints;
 	for (unsigned int i = 0; i < count; i++)
 	{
 		long ix = (m_doublePrecision) ? readS32() : readS16();
 		long iy = (m_doublePrecision) ? readS32() : readS16();
-		TRANSFORM_XY(ix,iy);
+		transformXY(ix,iy);
 
 		long ax = (m_doublePrecision) ? readS32() : readS16();
 		long ay = (m_doublePrecision) ? readS32() : readS16();
-		TRANSFORM_XY(ax,ay);
+		transformXY(ax,ay);
 
 		long tx = (m_doublePrecision) ? readS32() : readS16();
 		long ty = (m_doublePrecision) ? readS32() : readS16();
-		TRANSFORM_XY(tx,ty);
+		transformXY(tx,ty);
 
-		element.insert("svg:x", (TO_DOUBLE(ax)/m_xres));
-		element.insert("svg:y", (TO_DOUBLE(ay)/m_yres));
+		element.insert("svg:x", (toDouble(ax)/m_xres));
+		element.insert("svg:y", (toDouble(ay)/m_yres));
 		if (i == 0)
 			element.insert("librevenge:path-action", "M");
 		else
 		{
-			element.insert("svg:x2", (TO_DOUBLE(ix)/m_xres));
-			element.insert("svg:y2", (TO_DOUBLE(iy)/m_yres));
+			element.insert("svg:x2", (toDouble(ix)/m_xres));
+			element.insert("svg:y2", (toDouble(iy)/m_yres));
 			element.insert("librevenge:path-action", "C");
 		}
 		path.append(element);
-		element.insert("svg:x1", (TO_DOUBLE(tx)/m_xres));
-		element.insert("svg:y1", (TO_DOUBLE(ty)/m_yres));
+		element.insert("svg:x1", (toDouble(tx)/m_xres));
+		element.insert("svg:y1", (toDouble(ty)/m_yres));
 	}
 
 	element.clear();
@@ -1623,11 +1644,11 @@ void WPG2Parser::handleRectangle()
 
 	long x1 = (m_doublePrecision) ? readS32() : readS16();
 	long y1 = (m_doublePrecision) ? readS32() : readS16();
-	TRANSFORM_XY(x1,y1);
+	transformXY(x1,y1);
 
 	long x2 = (m_doublePrecision) ? readS32() : readS16();
 	long y2 = (m_doublePrecision) ? readS32() : readS16();
-	TRANSFORM_XY(x2,y2);
+	transformXY(x2,y2);
 
 	long xs1 = (x1 <= x2) ? x1 : x2;
 	long xs2 = (x1 <= x2) ? x2 : x1;
@@ -1637,14 +1658,14 @@ void WPG2Parser::handleRectangle()
 	long rx = (m_doublePrecision) ? readS32() : readS16();
 	long ry = (m_doublePrecision) ? readS32() : readS16();
 
-	::librevenge::RVNGPropertyList propList;
-	propList.insert("svg:x", (TO_DOUBLE(xs1) / m_xres));
-	propList.insert("svg:width", (TO_DOUBLE(xs2-xs1) / m_xres));
-	propList.insert("svg:y", (TO_DOUBLE(ys1) / m_yres));
-	propList.insert("svg:height", (TO_DOUBLE(ys2-ys1) / m_yres));
+	librevenge::RVNGPropertyList propList;
+	propList.insert("svg:x", (toDouble(xs1) / m_xres));
+	propList.insert("svg:width", (toDouble(xs2-xs1) / m_xres));
+	propList.insert("svg:y", (toDouble(ys1) / m_yres));
+	propList.insert("svg:height", (toDouble(ys2-ys1) / m_yres));
 
-	propList.insert("svg:rx", (TO_DOUBLE(rx)/m_xres));
-	propList.insert("svg:ry", (TO_DOUBLE(ry)/m_yres));
+	propList.insert("svg:rx", (toDouble(rx)/m_xres));
+	propList.insert("svg:ry", (toDouble(ry)/m_yres));
 
 	if (objCh.filled || m_gradient.count())
 		tmpStyle.insert("svg:linearGradient", m_gradient);
@@ -1690,9 +1711,9 @@ void WPG2Parser::handleArc()
 	long ey = (m_doublePrecision) ? readS32() : readS16();
 	ey += cy;
 
-	TRANSFORM_XY(cx,cy);
-	TRANSFORM_XY(ix,iy);
-	TRANSFORM_XY(ex,ey);
+	transformXY(cx,cy);
+	transformXY(ix,iy);
+	transformXY(ex,ey);
 
 	if (objCh.filled || m_gradient.count())
 		tmpStyle.insert("svg:linearGradient", m_gradient);
@@ -1700,11 +1721,11 @@ void WPG2Parser::handleArc()
 
 	if (ix == ex && iy == ey)
 	{
-		::librevenge::RVNGPropertyList propList;
-		propList.insert("svg:cx", (TO_DOUBLE(cx) / m_xres));
-		propList.insert("svg:cy", (TO_DOUBLE(cy) / m_xres));
-		propList.insert("svg:rx", (TO_DOUBLE(radx) / m_xres));
-		propList.insert("svg:ry", (TO_DOUBLE(rady) / m_xres));
+		librevenge::RVNGPropertyList propList;
+		propList.insert("svg:cx", (toDouble(cx) / m_xres));
+		propList.insert("svg:cy", (toDouble(cy) / m_xres));
+		propList.insert("svg:rx", (toDouble(radx) / m_xres));
+		propList.insert("svg:ry", (toDouble(rady) / m_xres));
 		if (objCh.rotate)
 			propList.insert("librevenge:rotate", objCh.rotationAngle, librevenge::RVNG_GENERIC);
 
@@ -1712,20 +1733,20 @@ void WPG2Parser::handleArc()
 	}
 	else
 	{
-		::librevenge::RVNGPropertyList element;
-		::librevenge::RVNGPropertyListVector path;
+		librevenge::RVNGPropertyList element;
+		librevenge::RVNGPropertyListVector path;
 
 		element.insert("librevenge:path-action", "M");
-		element.insert("svg:x", (TO_DOUBLE(ix)/m_xres));
-		element.insert("svg:y", (TO_DOUBLE(iy)/m_yres));
+		element.insert("svg:x", (toDouble(ix)/m_xres));
+		element.insert("svg:y", (toDouble(iy)/m_yres));
 		path.append(element);
 		element.clear();
 
 		element.insert("librevenge:path-action", "A");
-		element.insert("svg:rx", (TO_DOUBLE(radx)/m_xres));
-		element.insert("svg:ry", (TO_DOUBLE(rady)/m_yres));
-		element.insert("svg:x", (TO_DOUBLE(ex)/m_xres));
-		element.insert("svg:y", (TO_DOUBLE(ey)/m_yres));
+		element.insert("svg:rx", (toDouble(radx)/m_xres));
+		element.insert("svg:ry", (toDouble(rady)/m_yres));
+		element.insert("svg:x", (toDouble(ex)/m_xres));
+		element.insert("svg:y", (toDouble(ey)/m_yres));
 		if (objCh.rotate)
 			element.insert("librevenge:rotate", objCh.rotationAngle, librevenge::RVNG_GENERIC);
 		path.append(element);
@@ -1765,8 +1786,8 @@ void WPG2Parser::handleBitmap()
 	long y1 = (m_doublePrecision) ? readS32() : readS16();
 	long x2 = (m_doublePrecision) ? readS32() : readS16();
 	long y2 = (m_doublePrecision) ? readS32() : readS16();
-	TRANSFORM_XY(x1,y1);
-	TRANSFORM_XY(x2,y2);
+	transformXY(x1,y1);
+	transformXY(x2,y2);
 
 	long hres = (m_doublePrecision) ? readS32() : readS16();
 	long vres = (m_doublePrecision) ? readS32() : readS16();
@@ -1776,10 +1797,10 @@ void WPG2Parser::handleBitmap()
 	long ys1 = (y1 <= y2) ? y1 : y2;
 	long ys2 = (y1 <= y2) ? y2 : y1;
 
-	m_bitmap.x1 = TO_DOUBLE(xs1) / m_xres;
-	m_bitmap.y1 = TO_DOUBLE(ys1) / m_yres;
-	m_bitmap.x2 = TO_DOUBLE(xs2) / m_xres;
-	m_bitmap.y2 = TO_DOUBLE(ys2) / m_yres;
+	m_bitmap.x1 = toDouble(xs1) / m_xres;
+	m_bitmap.y1 = toDouble(ys1) / m_yres;
+	m_bitmap.x2 = toDouble(xs2) / m_xres;
+	m_bitmap.y2 = toDouble(ys2) / m_yres;
 	if (hres == 0)
 		hres = 72;
 	m_bitmap.hres = hres;
@@ -2008,7 +2029,7 @@ void WPG2Parser::handleBitmapData()
 
 	// prepare the bitmap structure for the listener
 	libwpg::WPGBitmap bitmap((int) width, (int) height, m_vFlipped, m_hFlipped);
-	::librevenge::RVNGPropertyList propList;
+	librevenge::RVNGPropertyList propList;
 	propList.insert("svg:x", (double)m_bitmap.x1);
 	propList.insert("svg:y", (double)m_bitmap.y1);
 	propList.insert("svg:width", (m_bitmap.x2 - m_bitmap.x1));
@@ -2099,7 +2120,7 @@ void WPG2Parser::handleBitmapData()
 		for (unsigned long y = 0; y < height; y++)
 			for (unsigned long x = 0; x < width; x++, i++)
 			{
-				unsigned short tmpColor = (unsigned short)(buffer[2*i] | (buffer[2*i+1] << 8));
+				auto tmpColor = (unsigned short)(buffer[2*i] | (buffer[2*i+1] << 8));
 				const libwpg::WPGColor color((tmpColor >> 10) & 0x1f, (tmpColor >> 5) & 0x1f, tmpColor & 0x1f);
 				bitmap.setPixel((int) x, (int) y, color);
 			}
@@ -2114,7 +2135,7 @@ void WPG2Parser::handleBitmapData()
 		for (unsigned long y = 0; y < height; y++)
 			for (unsigned long x = 0; x < width; x++, i++)
 			{
-				unsigned short tmpColor = (unsigned short)(buffer[2*i] | (buffer[2*i+1] << 8));
+				auto tmpColor = (unsigned short)(buffer[2*i] | (buffer[2*i+1] << 8));
 				const libwpg::WPGColor color(tmpColor & 0x0f, (tmpColor >> 4) & 0x0f, (tmpColor >> 8) & 0x0f, (tmpColor >> 12) & 0x0f);
 				bitmap.setPixel((int) x, (int) y, color);
 			}
@@ -2122,26 +2143,28 @@ void WPG2Parser::handleBitmapData()
 		propList.insert("office:binary-data", bitmap.getDIB());
 		m_painter->drawGraphicObject(propList);
 	}
+	// format 8: RGB, with 8 bits per colour
 	else if (color_format == 8)
 	{
 		size_t i = 0;
 		for (unsigned long y = 0; y < height; y++)
 			for (unsigned long x = 0; x < width; x++, i++)
 			{
-				const libwpg::WPGColor color(buffer[3*i+2] & 0xff, buffer[3*i+1] & 0xff, buffer[3*i] & 0xff);
+				const libwpg::WPGColor color(buffer[3*i] & 0xff, buffer[3*i+1] & 0xff, buffer[3*i+2] & 0xff);
 				bitmap.setPixel((int) x, (int) y, color);
 			}
 
 		propList.insert("office:binary-data", bitmap.getDIB());
 		m_painter->drawGraphicObject(propList);
 	}
+	// format 9: RGBA, with 8 bits per colour
 	else if (color_format == 9)
 	{
 		size_t i = 0;
 		for (unsigned long y = 0; y < height; y++)
 			for (unsigned long x = 0; x < width; x++, i++)
 			{
-				const libwpg::WPGColor color(buffer[4*i+3], buffer[4*i+2] & 0xff, buffer[4*i+1] & 0xff, buffer[4*i] & 0xff);
+				const libwpg::WPGColor color(buffer[4*i], buffer[4*i+1] & 0xff, buffer[4*i+2] & 0xff, buffer[4*i+3] & 0xff);
 				bitmap.setPixel((int) x, (int) y, color);
 			}
 
@@ -2177,18 +2200,18 @@ void WPG2Parser::handleObjectCapsule()
 	long y1 = (m_doublePrecision) ? readS32() : readS16();
 	long x2 = (m_doublePrecision) ? readS32() : readS16();
 	long y2 = (m_doublePrecision) ? readS32() : readS16();
-	TRANSFORM_XY(x1,y1);
-	TRANSFORM_XY(x2,y2);
+	transformXY(x1,y1);
+	transformXY(x2,y2);
 
 	long xs1 = (x1 <= x2) ? x1 : x2;
 	long xs2 = (x1 <= x2) ? x2 : x1;
 	long ys1 = (y1 <= y2) ? y1 : y2;
 	long ys2 = (y1 <= y2) ? y2 : y1;
 
-	m_binaryData.x1 = TO_DOUBLE(xs1) / m_xres;
-	m_binaryData.y1 = TO_DOUBLE(ys1) / m_yres;
-	m_binaryData.x2 = TO_DOUBLE(xs2) / m_xres;
-	m_binaryData.y2 = TO_DOUBLE(ys2) / m_yres;
+	m_binaryData.x1 = toDouble(xs1) / m_xres;
+	m_binaryData.y1 = toDouble(ys1) / m_yres;
+	m_binaryData.x2 = toDouble(xs2) / m_xres;
+	m_binaryData.y2 = toDouble(ys2) / m_yres;
 
 	unsigned long numDescriptions = readU16();
 
@@ -2264,7 +2287,7 @@ void WPG2Parser::handleObjectImage()
 	unsigned accessoryDataLength = readU16();
 	m_input->seek((long) accessoryDataLength, librevenge::RVNG_SEEK_CUR);
 
-	::librevenge::RVNGPropertyList propList;
+	librevenge::RVNGPropertyList propList;
 	propList.insert("svg:x", (double)m_binaryData.x1);
 	propList.insert("svg:y", (double)m_binaryData.y1);
 	propList.insert("svg:width", (m_binaryData.x2 - m_binaryData.x1));
@@ -2273,9 +2296,13 @@ void WPG2Parser::handleObjectImage()
 
 	WPG_DEBUG_MSG(("Image Object Mime Type : %s\n", propList["librevenge:mime-type"]->getStr().cstr()));
 
-	::librevenge::RVNGBinaryData binaryData;
-	while (!m_input->isEnd() && m_input->tell() <= m_recordEnd)
-		binaryData.append((unsigned char)readU8());
+	librevenge::RVNGBinaryData binaryData;
+	if (!m_input->isEnd() && m_input->tell() <= m_recordEnd)
+	{
+		unsigned long numBytes = 0;
+		const unsigned char *const data = m_input->read(m_recordEnd - m_input->tell(), numBytes);
+		binaryData = librevenge::RVNGBinaryData(data, numBytes);
+	}
 	WPG_DEBUG_MSG(("     Image Object Size : %li\n", (unsigned long)binaryData.size()));
 
 	// temporary for debug - dump the binary data (need to have write access in the current directory
@@ -2310,7 +2337,7 @@ void WPG2Parser::handleTextLine()
 	unsigned short textFlags = readU16();
 	long x = (m_doublePrecision) ? readS32() : readS16();
 	long y = (m_doublePrecision) ? readS32() : readS16();
-	TRANSFORM_XY(x,y);
+	transformXY(x,y);
 
 	unsigned char horizontalAlignment = readU8();
 	unsigned char verticalAlignment = readU8();
@@ -2324,10 +2351,10 @@ void WPG2Parser::handleTextLine()
 	WPG_DEBUG_MSG(("    vertical alignment : 0x%.2x\n", verticalAlignment));
 	WPG_DEBUG_MSG(("        baseline angle : %.4f\n", baseLineAngle));
 
-	m_textData.x1 = TO_DOUBLE(x)/m_xres;
-	m_textData.y1 = TO_DOUBLE(y)/m_yres;
-	m_textData.x2 = TO_DOUBLE(x)/m_xres;
-	m_textData.y2 = TO_DOUBLE(y)/m_yres;
+	m_textData.x1 = toDouble(x)/m_xres;
+	m_textData.y1 = toDouble(y)/m_yres;
+	m_textData.x2 = toDouble(x)/m_xres;
+	m_textData.y2 = toDouble(y)/m_yres;
 	m_textData.flags = textFlags;
 	m_textData.horAlign = horizontalAlignment;
 	m_textData.vertAlign = verticalAlignment;
@@ -2348,8 +2375,8 @@ void WPG2Parser::handleTextBlock()
 	long y1 = (m_doublePrecision) ? readS32() : readS16();
 	long x2 = (m_doublePrecision) ? readS32() : readS16();
 	long y2 = (m_doublePrecision) ? readS32() : readS16();
-	TRANSFORM_XY(x1,y1);
-	TRANSFORM_XY(x2,y2);
+	transformXY(x1,y1);
+	transformXY(x2,y2);
 
 	long xs1 = (x1 <= x2) ? x1 : x2;
 	long xs2 = (x1 <= x2) ? x2 : x1;
@@ -2361,10 +2388,10 @@ void WPG2Parser::handleTextBlock()
 	WPG_DEBUG_MSG(("             x2 : %li\n", xs2));
 	WPG_DEBUG_MSG(("             y2 : %li\n", ys2));
 
-	m_textData.x1 = TO_DOUBLE(xs1)/m_xres;
-	m_textData.y1 = TO_DOUBLE(ys1)/m_yres;
-	m_textData.x2 = TO_DOUBLE(xs2)/m_xres;
-	m_textData.y2 = TO_DOUBLE(ys2)/m_yres;
+	m_textData.x1 = toDouble(xs1)/m_xres;
+	m_textData.y1 = toDouble(ys1)/m_yres;
+	m_textData.x2 = toDouble(xs2)/m_xres;
+	m_textData.y2 = toDouble(ys2)/m_yres;
 	m_textData.flags = 0;
 	m_textData.horAlign = 0;
 	m_textData.vertAlign = 0;
@@ -2387,12 +2414,15 @@ void WPG2Parser::handleTextData()
 		return;
 	if (!m_drawTextData)
 		return;
+	if (m_input->isEnd() || m_input->tell() >= m_recordEnd)
+		return;
+
+	unsigned long numBytes = 0;
+	const unsigned char *const data = m_input->read(m_recordEnd - m_input->tell(), numBytes);
+	librevenge::RVNGBinaryData textData(data, numBytes);
 
-	::librevenge::RVNGBinaryData textData;
-	while (!m_input->isEnd() && m_input->tell() <= m_recordEnd)
-		textData.append((unsigned char)readU8());
 	WPGTextDataHandler handler(m_painter);
-	::librevenge::RVNGPropertyList propList;
+	librevenge::RVNGPropertyList propList;
 
 	propList.insert("svg:x", (double)m_textData.x1);
 	propList.insert("svg:y", (double)m_textData.y1);
@@ -2439,4 +2469,17 @@ bool WPG2Parser::checkRLESize(const unsigned bytes) const
 	return bytes < maxSize;
 }
 
+double WPG2Parser::toDouble(long x) const
+{
+	return m_doublePrecision ? double(x)/65536.0 : double(x);
+}
+
+void WPG2Parser::transformXY(long &x, long &y) const
+{
+	m_matrix.transform(x,y);
+	x = safeSubtract(x, m_xofs);
+	y = safeSubtract(y, m_yofs);
+	y = safeSubtract(m_height, y);
+}
+
 /* vim:set shiftwidth=4 softtabstop=4 noexpandtab: */
